/**
      This file is part of Adguard for iOS (https://github.com/AdguardTeam/AdguardForiOS).
      Copyright © Adguard Software Limited. All rights reserved.

      Adguard for iOS is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 3 of the License, or
      (at your option) any later version.

      Adguard for iOS is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with Adguard for iOS.  If not, see <http://www.gnu.org/licenses/>.
*/

import Foundation

/** Wraps the observer token received from
 NotificationCenter.addObserver(forName:object:queue:using:)
 and unregisters it in deinit.
 We shoul do it because "block based observers via the -[NSNotificationCenter addObserver​ForName:​object:​queue:​usingBlock:] method still need to be un-registered"
 https://developer.apple.com/library/archive/releasenotes/Foundation/RN-FoundationOlderNotes/index.html#10_11NotificationCenter
 */
final class NotificationToken: NSObject {
    let notificationCenter: NotificationCenter
    let token: Any

    init(notificationCenter: NotificationCenter = .default, token: Any) {
        self.notificationCenter = notificationCenter
        self.token = token
    }

    deinit {
        notificationCenter.removeObserver(token)
    }
}

final class ObserverToken: NSObject {
    let observableObject: NSObject
    unowned let observer: NSObject
    let keyPath: String
    
    init(observableObject: NSObject, observer: NSObject, keyPath: String) {
        self.observableObject = observableObject
        self.observer = observer
        self.keyPath = keyPath
    }

    deinit {
        observableObject.removeObserver(observer, forKeyPath: keyPath)
    }
}


extension NotificationCenter {
    /// Convenience wrapper for addObserver(forName:object:queue:using:)
    /// that returns our custom NotificationToken.
    func observe(name: NSNotification.Name?, object obj: Any?,
        queue: OperationQueue?, using block: @escaping (Notification) -> ())
        -> NotificationToken
    {
        let token = addObserver(forName: name, object: obj, queue: queue, using: block)
        return NotificationToken(notificationCenter: self, token: token)
    }
}

extension NSObject {
    func addObseverWithToken(_ observer: NSObject, keyPath: String, options: NSKeyValueObservingOptions, context: UnsafeMutableRawPointer?)->ObserverToken {
        addObserver(observer, forKeyPath: keyPath, options: options, context: context)
        return ObserverToken(observableObject: self, observer: observer, keyPath: keyPath)
    }
}
